from __future__ import print_function
import ctypes
from .var import fVar, fParam
import numpy as np
from .utils import *
from .fnumpy import *
from .errors import *

_index_t = ctypes.c_int64
_size_t = ctypes.c_int64

# Pre generate alloc array descriptors
class _bounds(ctypes.Structure):
    _fields_=[("stride",_index_t),
              ("lbound",_index_t),
              ("ubound",_index_t)]

class _fAllocArray1D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*1)
              ]

class _fAllocArray2D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*2)
              ]
              
class _fAllocArray3D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*3)
              ]
              
class _fAllocArray4D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*4)
              ]
class _fAllocArray5D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*5)
              ]
              
class _fAllocArray6D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*6)
              ]
              
class _fAllocArray7D(ctypes.Structure):
    _fields_=[('base_addr',ctypes.c_void_p), 
              ('offset',_size_t), 
              ('dtype',_index_t),
              ('dims',_bounds*7)
              ]

# None is in there so we can do 1 based indexing
_listFAllocArrays=[None,_fAllocArray1D,_fAllocArray2D,_fAllocArray3D,
                    _fAllocArray4D,_fAllocArray5D,_fAllocArray6D,
                    _fAllocArray7D] 

              

class fExplicitArray(fVar):

    def __init__(self, lib, obj):
        self.__dict__.update(obj)
        self._lib = lib
        self._pytype = np.array
        self.ctype = self.var['array']['ctype']
        
        self._ctype = self.ctype_def()
        
        if 'array' in self.var:
          self.__dict__.update(obj['var'])
        
        self.ndims = int(self.array['ndims'])
        #self._ctype_f = self.ctype_def_func()
        self._dtype=self.pytype+str(8*ctypes.sizeof(self._ctype))

        #Store the ref to the lib object
        try:   
            self._ref = self._get_from_lib()
        except NotInLib:
            self._ref = None

    def ctype_to_py(self, value):
        """
        Pass in a ctype value returns the python representation of it
        """
        return self._get_var_by_iter(value, self._array_size())
        
    def py_to_ctype_f(self, value):
        """
        Pass in a python value returns the ctype representation of it, 
        suitable for a function
        
        Second return value is anything that needs to go at the end of the
        arg list, like a string len
        """
        self._data = np.asfortranarray(value.T.astype(self._dtype))

        return self._data,None
        
    def ctype_to_py_f(self, value):
        """
        Pass in a ctype value returns the python representation of it,
        as returned by a function (may be a pointer)
        """
        self._value = np.asfortranarray(value,dtype=self._dtype)
        return self._value

    def pytype_def(self):
        return self._pytype

    def ctype_def(self):
        """
        The ctype type of this object
        """
        if '_cached_ctype' not in self.__dict__:
            self._cached_ctype = getattr(ctypes, self.ctype)
        
        return self._cached_ctype

    def ctype_def_func(self,pointer=False,intent=''):
        """
        The ctype type of a value suitable for use as an argument of a function

        May just call ctype_def
        
        Second return value is anythng that needs to go at the end of the
        arg list, like a string len
        """
        if pointer:
            raise ValueError("Cant have explicit array as a pointer")
        
        x=np.ctypeslib.ndpointer(dtype=self._dtype,ndim=self.ndims,
                                flags='F_CONTIGUOUS')
        y=None
        return x,y        
        
    def set_mod(self, value):
        """
        Set a module level variable
        """
        v = value.flatten(order='C')
        self._set_var_from_iter(self._ref, v, self._array_size())
        
    def get(self,copy=True):
        """
        Get a module level variable
        """
        s = self.ctype_to_py(self._ref)
        shape = self._make_array_shape()
        return np.reshape(s, shape)

    def _make_array_shape(self,bounds=None):
        if bounds is None:
            bounds = self.array['bounds']
        
        shape = []
        for i, j in zip(bounds[0::2], bounds[1::2]):
            shape.append(j - i + 1)
        return shape

    def _array_size(self,bounds=None):
        return np.product(self._make_array_shape(bounds))
       
    def py_to_ctype_p(self,value):
        """
        The ctype represnation suitable for function arguments wanting a pointer
        """

        raise AttributeError("Cant have explicit array as a pointer")


class fDummyArray(fVar):
    _GFC_MAX_DIMENSIONS = 7

    _GFC_DTYPE_RANK_MASK = 0x07
    _GFC_DTYPE_TYPE_SHIFT = 3
    _GFC_DTYPE_TYPE_MASK = 0x38
    _GFC_DTYPE_SIZE_SHIFT = 6

    _BT_UNKNOWN = 0
    _BT_INTEGER = _BT_UNKNOWN + 1
    _BT_LOGICAL = _BT_INTEGER + 1
    _BT_REAL = _BT_LOGICAL + 1
    _BT_COMPLEX = _BT_REAL + 1
    _BT_DERIVED = _BT_COMPLEX + 1
    _BT_CHARACTER = _BT_DERIVED + 1
    _BT_CLASS = _BT_CHARACTER + 1
    _BT_PROCEDURE = _BT_CLASS + 1
    _BT_HOLLERITH = _BT_PROCEDURE + 1
    _BT_VOID = _BT_HOLLERITH + 1
    _BT_ASSUMED = _BT_VOID + 1


    def __init__(self, lib, obj):
        self.__dict__.update(obj)
        self._lib = lib

        if 'array' in self.var:
          self.__dict__.update(obj['var'])

        self.ndim = int(self.array['ndims'])
        self._lib = lib
        
        self._desc = self._setup_desc()
        self._ctype_single = getattr(ctypes,self.ctype)
        #self._ctype = getattr(ctypes,self.ctype)
        self._ctype = self._desc
        self._ctype_desc = ctypes.POINTER(self._desc)
        self.npdtype=self.pytype+str(8*ctypes.sizeof(self._ctype_single))
        

    def _setup_desc(self):
        return _listFAllocArrays[self.ndim]

    def _get_pointer(self):
        return self._ctype_desc.from_address(ctypes.addressof(getattr(self._lib,self.mangled_name)))

    def set_mod(self, value):
        """
        Set a module level variable
        """
        
        if not self.ndim == value.ndim:
            raise ValueError("Array size mismatch "+str(self.ndim)+' '+str(value.ndim))
        
        
        self._value = value.astype(self.npdtype)
        
        #Did we make a copy?
        if self._id(self._value)==self._id(value):
            remove_ownership(value)
        remove_ownership(self._value)
        
        p = self._get_pointer()
        if p:
            self._set_to_pointer(self._value,p.contents)
        
        return 
        
    def set_func_arg(self,value):
        #Create an allocatable array
        self._value_array = self._desc()

        self._value = np.asfortranarray(value).astype(self.npdtype)
        self._set_to_pointer(self._value,self._value_array)
        
        
    def _set_to_pointer(self,value,p):
        if value.ndim > self._GFC_MAX_DIMENSIONS:
            raise ValueError("Array too big")
        
        p.base_addr = value.ctypes.get_data()
        p.offset = _size_t(-1)
        
        p.dtype = self._get_dtype()
        
        for i in range(self.ndim):
            p.dims[i].stride = _index_t(value.strides[i]//ctypes.sizeof(self._ctype_single))
            p.dims[i].lbound = _index_t(1)
            p.dims[i].ubound = _index_t(value.shape[i])
            
        return

    def get(self,copy=False):
        """
        Get a module level variable
        """
        if self._func_arg:
            return self._value
           
        p = self._get_pointer()
        return self._get_from_pointer(p.contents,copy)
        
    def _get_from_pointer(self,p,copy=False):
        base_addr = p.base_addr
        if not self._isallocated():
            return np.zeros(1)
            #raise ValueError("Array not allocated yet")
        
        offset = p.offset
        dtype = p.dtype
        
        dims=[]
        shape=[]
        for i in range(self.ndim):
            dims.append({})
            dims[i]['stride'] = p.dims[i].stride
            dims[i]['lbound'] = p.dims[i].lbound
            dims[i]['ubound'] = p.dims[i].ubound
            
        for i in range(self.ndim):
            shape.append(dims[i]['ubound']-dims[i]['lbound']+1)
            
        self._shape=tuple(shape)
        size = np.product(shape)
        
        if copy:
            # When we want a copy of the array not a pointer to the fortran memoray
            res = self._get_var_from_address(base_addr,size=size)
            res = np.asfortranarray(res)
            res = res.reshape(shape).astype(self.npdtype)
        else:
            # When we want to pointer to the underlaying fortran memoray
            # will leak as we dont have a deallocate call to call in a del func
            ptr = ctypes.cast(base_addr,ctypes.POINTER(self._ctype_single))
            res = np.ctypeslib.as_array(ptr,shape= self._shape)
        
        return res
        

    def py_to_ctype(self, value):
        """
        Pass in a python value returns the ctype representation of it
        """
        self.set_func_arg(value)
        return self._value_array
        
    def py_to_ctype_f(self, value):
        """
        Pass in a python value returns the ctype representation of it, 
        suitable for a function
        
        Second return value is anything that needs to go at the end of the
        arg list, like a string len
        """
        return self.py_to_ctype(value),None

    def ctype_to_py(self, value):
        """
        Pass in a ctype value returns the python representation of it
        """
        return self.ctype_to_py_f(value.contents)
        
    def ctype_to_py_f(self, value):
        """
        Pass in a ctype value returns the python representation of it,
        as returned by a function (may be a pointer)
        """
        if hasattr(value,'contents'):
            return self._get_from_pointer(value.contents)
        else:
            return self._get_from_pointer(value)

            
    def py_to_ctype_p(self,value):
        """
        The ctype represnation suitable for function arguments wanting a pointer
        """
        return self.py_to_ctype(value)
            

    def pytype_def(self):
        return np.array

    def ctype_def(self):
        """
        The ctype type of this object
        """
        return self._ctype_desc

    def ctype_def_func(self,pointer=False,intent=''):
        """
        The ctype type of a value suitable for use as an argument of a function

        May just call ctype_def
        
        Second return value is anythng that needs to go at the end of the
        arg list, like a string len
        """

        return self.ctype_def(),None

    def _get_dtype(self):
        ftype=self._get_ftype()
        d=self.ndim
        d=d|(ftype<<self._GFC_DTYPE_TYPE_SHIFT)
        d=d|(ctypes.sizeof(self._ctype_single)<<self._GFC_DTYPE_SIZE_SHIFT)
        return d

    def _get_ftype(self):
        ftype=None
        dtype=self.ctype
        if 'c_int' in dtype:
            ftype=self._BT_INTEGER
        elif 'c_double' in dtype or 'c_real' in dtype or 'c_float' in dtype:
            ftype=self._BT_REAL
        elif 'c_bool' in dtype:
            ftype=self._BT_LOGICAL
        elif 'c_char' in dtype:
            ftype=self._BT_CHARACTER
        else:
            raise ValueError("Cant match dtype, got "+dtype)
        return ftype

    def __str__(self):
        x=self.get()
        if x is None:
            return "<array>"
        else:
            return str(self.get())
        
    def __repr__(self):
        x=self.get()
        if x is None:
            return "<array>"
        else:
            return repr(self.get())

    def __getattr__(self, name): 
        if name in self.__dict__:
            return self.__dict__[name]

        #return getattr(self.get(),name)
        
    def __del__(self):
        if '_value' in self.__dict__:
            #Problem occurs as both fortran and numpy are pointing to same memory address
            #Thus if fortran deallocates the array numpy will try to free the pointer
            #when del is called casuing a double free error
            
            #By calling remove_ownership we tell numpy it dosn't own the data
            #thus is shouldn't call free(ptr).
            remove_ownership(self._value)
            
            #Maybe leaks if fortran doesn't dealloc the array
                
                
    def _isallocated(self):
        
        try:
            p = self._get_pointer()
        except TypeError:
            return False
        if p.contents.base_addr:
            #Base addr is NULL if deallocated
            return True
        else:
            return False
        
    def _id(self,x):
        return x.ctypes.data
   
class fAssumedShape(fDummyArray):
    def _get_pointer(self):
        return self._ctype_desc.from_address(ctypes.addressof(self._value_array))
    
    
    def set_func_arg(self,value):
        
        super(fAssumedShape,self).set_func_arg(value)
        
        #Fix up bounds
    
        #From gcc source code
        #Parsed       Lower   Upper  Returned
        #------------------------------------
          #:           NULL    NULL   AS_DEFERRED (*)
          #x            1       x     AS_EXPLICIT
          #x:           x      NULL   AS_ASSUMED_SHAPE
          #x:y          x       y     AS_EXPLICIT
          #x:*          x      NULL   AS_ASSUMED_SIZE
          #*            1      NULL   AS_ASSUMED_SIZE
          
       # for i in range(self.ndim):
            #print(self._value_array.dims[i].lbound,self._value_array.dims[i].ubound)
            #self._value_array.dims[i].ubound=0
            #self._value_array.dims[i].lbound=0
            
    def __str__(self):
        return str(self._value_array)
        
    def __repr__(self):
        return repr(self._value_array)

    def py_to_ctype(self, value):
        """
        Pass in a python value returns the ctype representation of it
        """
        self.set_func_arg(value)
        return self._value_array
        
    def py_to_ctype_f(self, value):
        """
        Pass in a python value returns the ctype representation of it, 
        suitable for a function
        
        Second return value is anything that needs to go at the end of the
        arg list, like a string len
        """
        return self.py_to_ctype(value),None    
    
class fAssumedSize(fExplicitArray):
    pass
    
class fAllocatableArray(fDummyArray):
    def py_to_ctype(self, value):
        """
        Pass in a python value returns the ctype representation of it
        """
        self.set_func_arg(value)
        
        # self._value_array needs to be empty if the array is allocatable and not
        # allready allocated
        self._value_array.base_addr=ctypes.c_void_p(0)
        
        return self._value_array
        
    def py_to_ctype_f(self, value):
        """
        Pass in a python value returns the ctype representation of it, 
        suitable for a function
        
        Second return value is anything that needs to go at the end of the
        arg list, like a string len
        """
        return self.py_to_ctype(value),None   
        
    def ctype_to_py_f(self, value):
        """
        Pass in a ctype value returns the python representation of it,
        as returned by a function (may be a pointer)
        """
        shape=[]
        for i in value.dims:
            shape.append(i.ubound-i.lbound+1)
        shape=tuple(shape)
        
        p=ctypes.POINTER(self._ctype_single)
        res=ctypes.cast(value.base_addr,p)
        return np.ctypeslib.as_array(res,shape=shape)
    
class fParamArray(fParam):
    def get(self):
        """
        A parameters value is stored in the dict, as we cant access them
        from the shared lib.
        """
        return np.array(self.value, dtype=self.pytype)







from __future__ import print_function
import ctypes
from .var import fVar
from .errors import *

class fStr(fVar):

    def __init__(self, lib, obj):
        self.__dict__.update(obj)
        self._lib = lib
        self._ctype = self.ctype_def()
       # self._ctype_f = self.ctype_def_func()
        self._pytype = str
        
        self.char_len = self.var['len']

        #Store the ref to the lib object
        try:   
            self._ref = self._get_from_lib()
        except NotInLib:
            self._ref = None

    def py_to_ctype(self, value):
        """
        Pass in a python value returns the ctype representation of it
        """
        return ctypes.c_char(value.encode())
        
    def py_to_ctype_f(self, value):
        """
        Pass in a python value returns the ctype representation of it, 
        suitable for a function
        
        Second return value is anythng that needs to go at the end of the
        arg list, like a string len
        """
        x,y=self.ctype_def_func()
        
        return x(value.encode()),y(len(value))

    def ctype_to_py(self, value):
        """
        Pass in a ctype value returns the python representation of it
        """
        return self._get_var_by_iter(value, self.char_len)
        
    def ctype_def(self):
        return ctypes.c_char

    def ctype_def_func(self,pointer=False,intent=''):
        """
        The ctype type of a value suitable for use as an argument of a function

        May just call ctype_def
        
        Second return value is anything that needs to go at the end of the
        arg list, like a string len
        """
        f = ctypes.c_char_p
        if pointer:
            f = ctypes.POINTER(f)
        
        return f,ctypes.c_int
        
    def py_to_ctype_p(self,value):
        """
        The ctype represnation suitable for function arguments wanting a pointer
        """
        return ctypes.c_char_p(value.encode())
        
    def ctype_to_py_f(self, value):
        """
        Pass in a ctype value returns the python representation of it,
        as returned by a function (may be a pointer)
        """
        if hasattr(value,'contents'):
            r = value.contents.value
        elif hasattr(value,'value'):
            r = value.value
        else:
            r = value
            
        try:
            r = r.decode()
        except AttributeError:
            pass
        except UnicodeDecodeError:
            r = ''
            
        return r
            

    def set_mod(self, value):
        """
        Set a module level variable
        """
        self._set_var_from_iter(self._ref, value.encode(), self.char_len)

    def get(self,copy=True):
        """
        Get a module level variable
        """
        s = self.ctype_to_py(self._ref)
        if not copy:
            raise ValueError("Must copy a string")
        
        return ''.join([i.decode() for i in s])



from __future__ import print_function
import os
import select


def find_key_val(list_dicts, key, value):
    v = value.lower()
    for idx, i in enumerate(list_dicts):
        if i[key].lower() == v:
            return idx


def read_pipe(pipe_out):
    def more_data():
        r, _, _ = select.select([pipe_out], [], [], 0)
        return bool(r)
    out = b''
    while more_data():
        out += os.read(pipe_out, 1024)
    return out.decode()



from __future__ import print_function
try:
    import __builtin__
except ImportError:
    import builtins as __builtin__

import ctypes
import pickle
import numpy as np
import errno

#from .parseMod import run_and_save, hash_file, fpyname
from .cmplx import fComplex, fParamComplex
from .arrays import fExplicitArray, fDummyArray, fParamArray, fAssumedShape , fAssumedSize
from .functions import fFunc
from .strings import fStr
from .types import fDerivedType
from .utils import *
from .var import fVar, fParam
from .errors import *

from . import parseMod as pm

WARN_ON_SKIP=False

#https://gcc.gnu.org/onlinedocs/gcc-6.1.0/gfortran/Argument-passing-conventions.html

class fFort(object):

    def __init__(self, libname, ffile, rerun=False):
        self._lib = ctypes.CDLL(libname)
        self._all_names=[]
        self._libname = libname
        self._fpy = pm.fpyname(ffile)
        self._load_data(ffile, rerun)
        self._init()

    def _load_data(self, ffile, rerun=False):
        try:
            f = open(self._fpy, 'rb')
        # FileNotFoundError does not exist on Python < 3.3
        except (OSError, IOError) as e: 
            if e.errno != errno.ENOENT:
                raise
            pm.run(ffile,save=True)
        else:
            f.close()

        with open(self._fpy, 'rb') as f:
            self.version = pickle.load(f)
            if self.version == 1:
                self._mod_data = pickle.load(f)

                if self._mod_data["checksum"] != pm.hashFile(ffile) or rerun:
                    x = pm.run(ffile,save=True,unpack=True)
                    self._mod_data = x[0]
                    self._mod_vars = x[1]
                    self._param = x[2]
                    self._funcs = x[3]
                    self._dt_defs = x[4]
                else:
                    self._mod_vars = pickle.load(f)
                    self._param = pickle.load(f)
                    self._funcs = pickle.load(f)
                    self._dt_defs = pickle.load(f)

    def _init(self):
        self._listVars = []
        self._listParams = []
        self._listFuncs = []
        
        for i in self._mod_vars:
            self._all_names.append(i['name'])
        

        for i in self._param:
            self._all_names.append(i['name'])
            
        for i in self._funcs:
            self._all_names.append(i['name'])
            
        self._all_names = set(self._all_names)
            
        for i in self._dt_defs:
            i['name']=i['name'].lower().replace("'","")
        
        for i in self._mod_vars:
            self._init_var(i)

        for i in self._param:
            self._init_param(i)

    
        for i in self._funcs:
            if 'arg' in i and len(i['arg'])>0:
                for k in i['arg']:
                    if 'dt' in k['var']:
                        k['var']['dt']['name']=k['var']['dt']['name'].lower().replace("'","")
                        for j in self._dt_defs:
                            if k['var']['dt']['name'].lower() == j['name'].lower():
                                k['_dt_def'] = j  

        # Must come last after the derived types are setup
        for i in self._funcs:
            self._init_func(i)

    def _init_var(self, obj):
        x=None
        if obj['var']['pytype'] == 'str':
            x = fStr(self._lib, obj)
        elif obj['var']['pytype'] == 'complex':
            x = fComplex(self._lib, obj)
        elif 'dt' in obj['var'] and obj['var']['dt']:
            x = fDerivedType(self._lib, obj,self._dt_defs)
        elif 'array' in obj['var']:
            array = obj['var']['array']['atype'] 
            if array == 'explicit':
                x = fExplicitArray(self._lib, obj)
            elif array == 'alloc' or array == 'pointer':
                x = fDummyArray(self._lib, obj)
            elif array == 'assumed_shape':
                x =  fAssumedShape(self._lib, obj)
            elif array == 'assumed_size':
                x = fAssumedSize(self._lib, obj)
        else:
            x = fVar(self._lib, obj)

        if x is not None:
            self.__dict__[x.name] = x
        else:
            print("Skipping init "+obj['name'])

    def _init_param(self, obj):
        if obj['param']['pytype']=='complex':
            x = fParamComplex(self._lib, obj)
        elif obj['param']['array']:
            x = fParamArray(self._lib, obj)
        else:
            x = fParam(self._lib, obj)

        self.__dict__[x.name] = x

    def _init_func(self, obj):
        x = fFunc(self._lib, obj,self._dt_defs)
        self.__dict__[x.name] = x

    def __getattr__(self, name):
        if name.lower() in self.__dict__:
            return self.__dict__[name.lower()]

        if '_all_names' in self.__dict__:
            if name.lower() in self._all_names:
                return self.__dict__[name.lower()].get()

        raise AttributeError("No variable " + name)

    def __setattr__(self, name, value):
        if '_all_names' in self.__dict__:
            if name in self._all_names:
                self.__dict__[name].set_mod(value)
                return
       
        self.__dict__[name] = value
        return



import sys
import numpy as np


def clean_list(l, idx):
    return [i for j, i in enumerate(l) if j not in idx]
    
def output(filename, *args):
    with open(filename, 'wb') as f:
        for i in args:
            pickle.dump(i, f, protocol=2)

    
#def split_brackets(value, remove_b=True):
    #'''
    #Split a string based on pairs of brackets, nested brackets are not split

    #Input:
        #'abc (def) (fgh () ())'

    #Outputs:
        #['abc (def')', '(fgh () ())']
    #'''
    #if remove_b:
        #if value.startswith('(') and value.endswith(')'):
            #value = value[1:-1]
    
    #res = []
    #start = False
    #count = 0
    #j = 0
    #for idx,i in enumerate(value):
        #if i == '(':
            #count = count + 1
            #start = True
        #if i == ')':
            #count = count - 1
        #if start:
            #if count == 0:
                #j2=idx+1
                #res.append(value[j:j2])
                #j=j2
                #start = False
                
    #if not j == len(value):
        #res.append(value[j:])
        
    #return res

def hextofloat(s):
    # Given hex like parameter '0.12decde@9' returns 5065465344.0
    man, exp =s.split('@')
    exp=int(exp)
    decimal = man.index('.')
    man = man[decimal+1:]
    man = man.ljust(exp,'0')
    man = man[:exp]+'.'+man[exp:]
    man = man +'P0'
    return float.fromhex(man)



from __future__ import print_function
import ctypes
from .var import fVar, fParam
from .errors import *

class fComplex(fVar):

    def __init__(self, lib, obj):
        self.__dict__.update(obj)
        self._lib = lib
        
        self.ctype=self.var['ctype']
        self.pytype=self.var['pytype']
        
        self._ctype = self.ctype_def()
        #self._ctype_f = self.ctype_def_func()
        self._pytype = self.pytype_def()
        
        #Store the ref to the lib object
        try:   
            self._ref = self._get_from_lib()
        except NotInLib:
            self._ref = None

    def py_to_ctype(self, value):
        """
        Pass in a python value returns the ctype representation of it
        """
        x = [value.real, value.imag]
        return self._set_var_from_iter(self._ref, x, 2)

    def ctype_to_py(self, value):
        """
        Pass in a ctype value returns the python representation of it
        """
        x = self._get_var_by_iter(value, 2)
        return self._pytype(x[0], x[1])

    def pytype_def(self):
        return complex

    def ctype_def(self):
        """
        The ctype type of this object
        """
        if '_cached_ctype' not in self.__dict__:
            self._cached_ctype = getattr(ctypes, self.ctype)
        
        return self._cached_ctype

    def set_mod(self, value):
        if isinstance(value, complex):
            self.py_to_ctype(value)
        else:
            raise ValueError("Not complex")

    def get(self,copy=True):
        s = self.ctype_to_py(self._ref)
        if not copy:
            raise ValueError("Must copy complex number")
        
        return s

    def __repr__(self):
        return str(self.get()) + " <complex>"


class fParamComplex(fParam):

    def get(self):
        """
        A parameters value is stored in the dict, as we cant access them 
        from the shared lib.
        """
        return self.value
